/*
 * This source code (including its associated software) is owned by
 * StorageCraft Technology Corporation and is protected by United States and
 * international intellectual property law, including copyright laws, patent
 * laws, and treaty provisions. Use of the source code without modification in
 * connection with StorageCraft products is permitted subject to the End-User
 * License Agreement for StorageCraft Products ("EULA"). A copy of the
 * StorageCraft EULA is available here:
 * https://www.storagecraft.com/legal/license-agreement. Any other use,
 * reproduction, or disclosure without the written permission of StorageCraft
 * Technology Corporation is prohibited.
 *
 * © 2014-2019 StorageCraft Technology Corporation. All rights reserved.
 */

#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/version.h>

#include "cow.h"
#include "uassert.h"
#include "debug.h"
#include "filter.h"
#include "proc.h"
#include "snapshot.h"
#include "snapshot_set.h"
#include "table.h"
#include "vdiff.h"

static struct mutex filter_lock;

void __init init_filters(void)
{
	int i;
	struct block_device_filter *table = stcvsm_filter_table();

	mutex_init(&filter_lock);
	for (i = 0; i < STCVSM_MAX_FILTERS; ++i) {
		init_waitqueue_head(&table[i].flush_waiters);
		atomic_set(&table[i].bdf_refs, 0);
		table[i].bdf_orig_reqfn = NULL;
	}
}

void shutdown_filters(void)
{
	int i;
	struct block_device_filter *table = stcvsm_filter_table();

	for (i = 0; i < STCVSM_MAX_FILTERS; ++i) {
		int vdiff_idx;

		/* Array is kept packed so shutdown backwards */
		for (vdiff_idx = table[i].num_vdiffs - 1; vdiff_idx >= 0; --vdiff_idx)
			shutdown_vdiff(table[i].vdiffs[vdiff_idx]);

		/* Currently never called when snapshots could exist */
		UASSERT(table[i].num_snapshots == 0, "Filter exists during shutdown");
	}
}

void dec_filtered_writes(struct block_device_filter *bdfilter)
{
	if (atomic_dec_and_test(&bdfilter->writes))
		wake_up_all(&bdfilter->flush_waiters);
}

static void wait_filtered_writes(struct block_device_filter *bdfilter)
{
	int r;
	wait_queue_head_t *wq = &bdfilter->flush_waiters;

	do {
		r = wait_event_interruptible(*wq, atomic_read(&bdfilter->writes) == 0);
	} while (r == -ERESTARTSYS);
}

int quiesce_filters(int quiet)
{
	int i;
	struct block_device_filter *table = stcvsm_filter_table();

	for (i = 0; i < STCVSM_MAX_FILTERS; ++i) {
		struct block_device_filter *bdfilter = &table[i];

		spin_lock(bdf_lock(bdfilter));
		if (quiet) {
			if (bdfilter->num_snapshots != 0) {
				spin_unlock(bdf_lock(bdfilter));
				return -EBUSY;
			}
			bdfilter->quiesced = 1;
		} else {
			bdfilter->quiesced = 0;
		}
		spin_unlock(bdf_lock(bdfilter));

		if (quiet)
			wait_filtered_writes(bdfilter);
	}
	if (quiet) {
		/* Quiesced, so refcount should be stable.  If module is still referenced,
		 * userspace is holding a snapshot open.
		 *
		 * Note that during unload, this has always been -1.  But kernels <= 3.18 were
		 * returning as unsigned long.  Commit d5db139ab in 3.19 returns as int, so receive
		 * as signed.
		 */
		long refs = module_refcount(THIS_MODULE);

		if (refs > 0) {
			stcvsm_warning(LOG_MODULE, "Quiesce failed; refcount is %ld\n", refs);
			return -EBUSY;
		}
	}
	return 0;
}

/** Get the parent block device. It is ok if the parent returned is itself to indicate the input
 * bdev is a "whole disk"
 */
struct block_device *get_parent_bdev(struct block_device *bdev)
{
	struct block_device *parent_bdev;

	parent_bdev = bdev->bd_contains;
	/* Check to see if stcvsm_open_blkdev or blkdev_get needs to be called */
	if (parent_bdev == NULL)
		return ERR_PTR(-ENODEV);
	/* Parent must be a "whole disk". For example, partitions of partitions aren't supported */
	if (parent_bdev->bd_contains != parent_bdev)
		return ERR_PTR(-EINVAL);
	parent_bdev = stcvsm_bdget(parent_bdev->bd_dev);
	if (!parent_bdev)
		return ERR_PTR(-ENOENT);  /* Or -ENOMEM?  We don't know. */
	return parent_bdev;
}

/** The original device's (hooked) request function.  All offsets are disk (not partition)
 * absolute.  Reads are passed through without interruption.  Writes may be filtered for
 * incremental tracking and copy-on-write.
 */
MAKE_REQUEST_RETTYPE block_device_filter_reqfn(struct request_queue *q, struct bio *bio,
		struct block_device_filter *bdfilter)
{
#ifdef MAKE_REQUEST_RETVAL
	MAKE_REQUEST_RETTYPE retval = MAKE_REQUEST_RETVAL;
#endif
	make_request_fn *reqfn = bdfilter->bdf_orig_reqfn;
	struct snapshot_context *snap = NULL;
	struct vdiff_context *vdiff = NULL;
	int dec_writes = 0;

	might_sleep();

	stcvsm_debug(LOG_FILTER, "[%llu] ORIG_REQFN [%s] sectors %u\n",
			(unsigned long long)bio_sector(bio),
			bio_data_dir(bio) == WRITE ? "write" : "read",
			bio_sectors(bio));

	spin_lock(bdf_lock(bdfilter));
	if (!atomic_read(&bdfilter->bdf_refs)) {
		/* IO was already in flight while bdfilter was getting destroyed.
		 * Since bdf_lock is held, q->make_request_fn should now be restored.
		 */
		reqfn = q->make_request_fn;
		spin_unlock(bdf_lock(bdfilter));
		goto do_req;
	}

	if (bio_data_dir(bio) == READ) {
		trace_orig_read_reqfn_io(MAJOR(bdfilter->bdf_bdev->bd_dev), MINOR(bdfilter->bdf_bdev->bd_dev),
				bio_sector(bio), bio_sectors(bio));
		spin_unlock(bdf_lock(bdfilter));
		goto do_req;
	}

	if (bio_sectors(bio)) {
		trace_orig_write_reqfn_io(MAJOR(bdfilter->bdf_bdev->bd_dev), MINOR(bdfilter->bdf_bdev->bd_dev),
				bio_sector(bio), bio_sectors(bio));
		vdiff = get_vdiff_bio(bdfilter, bio);
		snap = get_snapshot_bio(bdfilter, bio);
	} else {
		trace_orig_ctl_reqfn_io(MAJOR(bdfilter->bdf_bdev->bd_dev), MINOR(bdfilter->bdf_bdev->bd_dev),
				0, 0);
	}
	spin_unlock(bdf_lock(bdfilter));

	/* Expect COW writes to be sent directly to make_request_fn, not submit_bio */
	UASSERT(bio->bi_end_io != cow_bdev_write_endio, "Unexpected COW IO");

	/* Ensure all prior writes have been submitted to original make_request_fn before
	 * submitting the FLUSH.
	 */
	if (stc_bio_is_barrier(bio))
		wait_filtered_writes(bdfilter);
	if (bio_sectors(bio)) {
		if (vdiff) {
			set_vdiff_range(vdiff, bio_sector(bio) - vdiff->vd_sector_start, bio_sectors(bio));
			put_vdiff(vdiff);
		}

		if (snap) {
			if (!test_bit(SNAPSHOT_ERROR, &snap->sc_flags) &&
					!test_bit(SNAPSHOT_SHUTDOWN, &snap->sc_flags)) {
				atomic_inc(&bdfilter->writes);
				if (orig_write_filter(q, bio, bdfilter, snap))
					reqfn = NULL;
				else
					dec_writes = 1;
			}
			put_snapshot(snap);
		}
	}

do_req:
	if (reqfn) {
#ifdef MAKE_REQUEST_RETVAL
		retval = reqfn(q, bio);
#else
		reqfn(q, bio);
#endif
		if (dec_writes)
			dec_filtered_writes(bdfilter);
	}
	stcvsm_debug(LOG_FILTER, "[%llu] ORIG_REQFN [%s] Finished\n",
			(unsigned long long)bio_sector(bio),
			bio_data_dir(bio) == WRITE ? "write" : "read");
#ifdef MAKE_REQUEST_RETVAL
	return retval;
#endif
}

void activate_filter(struct block_device_filter *bdfilter)
{
	struct request_queue *q = bdfilter->bdf_disk->queue;

	spin_lock_irq(q->queue_lock);
	if (q->make_request_fn != bdfilter->bdf_reqfn) {
		stcvsm_debug(LOG_FILTER, "Replacing request function (%p->%p)\n",
				q->make_request_fn, bdfilter->bdf_reqfn);
		UASSERT(q->make_request_fn, "No 'make_request_fn' function value");
		bdfilter->bdf_orig_reqfn = q->make_request_fn;
		q->make_request_fn = bdfilter->bdf_reqfn;
	}
	spin_unlock_irq(q->queue_lock);
}

void deactivate_filter(struct block_device_filter *bdfilter)
{
	if (bdfilter->bdf_orig_reqfn) {
		struct request_queue *q = bdfilter->bdf_disk->queue;

		stcvsm_debug(LOG_FILTER, "Restoring request function (%p->%p)\n",
				bdfilter->bdf_orig_reqfn, q->make_request_fn);
		spin_lock_irq(q->queue_lock);
		q->make_request_fn = bdfilter->bdf_orig_reqfn;
		bdfilter->bdf_orig_reqfn = NULL;
		spin_unlock_irq(q->queue_lock);
	}
}

struct vdiff_context *get_filter_vdiff(struct block_device_filter *bdfilter, dev_t dev)
{
	int i;

	for (i = 0; i < bdfilter->num_vdiffs; ++i) {
		if (bdfilter->vdiffs[i]->vd_dev == dev)
			return bdfilter->vdiffs[i];
	}
	return NULL;
}

int attach_filter_vdiff(struct block_device_filter *bdfilter, struct vdiff_context *vdiff)
{
	if (bdfilter->quiesced)
		return -EBUSY;
	if (get_filter_vdiff(bdfilter, vdiff->vd_dev) != NULL)
		return -EEXIST;
	if (bdfilter->num_vdiffs >= MAX_VDIFFS_PER_FILTER)
		return -EMFILE;
	stcvsm_atomic_inc(&bdfilter->bdf_refs);
	stcvsm_debug(LOG_REFS, "bdf_refs %d\n", atomic_read(&bdfilter->bdf_refs));
	vdiff->bdfilter = bdfilter;
	bdfilter->vdiffs[bdfilter->num_vdiffs] = vdiff;
	stcvsm_debug(LOG_FILTER, "Increasing num_vdiffs from %d to %d\n", bdfilter->num_vdiffs,
			bdfilter->num_vdiffs + 1);
	++bdfilter->num_vdiffs;
	return 0;
}

void detach_filter_vdiff(struct block_device_filter *bdfilter, struct vdiff_context *vdiff)
{
	int idx, idx1;

	for (idx = 0; idx < bdfilter->num_vdiffs; ++idx) {
		if (bdfilter->vdiffs[idx] == vdiff) {
			stcvsm_debug(LOG_FILTER, "Decreasing num_vdiffs from %d to %d\n",
					bdfilter->num_vdiffs, bdfilter->num_vdiffs - 1);
			--bdfilter->num_vdiffs;
			/* Keep the array from getting fragmented */
			for (idx1 = idx; idx1 < bdfilter->num_vdiffs; ++idx1)
				bdfilter->vdiffs[idx1] = bdfilter->vdiffs[idx1 + 1];
			bdfilter->vdiffs[idx1] = NULL;
			break;
		}
	}
}

struct block_device_filter *get_existing_filter(dev_t dev)
{
	int i;
	struct block_device_filter *table = stcvsm_filter_table();

	/* Must grab each lock to avoid using partially constructed or destructed filter. */
	for (i = 0; i < STCVSM_MAX_FILTERS; ++i) {
		struct block_device_filter *bdfilter = &table[i];

		spin_lock(bdf_lock(bdfilter));
		if (atomic_read(&bdfilter->bdf_refs) &&
				bdfilter->bdf_bdev->bd_dev == dev) {
			stcvsm_info(LOG_FILTER, "(%d,%d) old %d refs %d\n", MAJOR(dev), MINOR(dev),
					i, atomic_read(&bdfilter->bdf_refs));
			stcvsm_atomic_inc(&bdfilter->bdf_refs);
			spin_unlock(bdf_lock(bdfilter));
			return bdfilter;
		}
		spin_unlock(bdf_lock(bdfilter));
	}
	return NULL;
}

static struct block_device_filter *get_parent_filter(struct block_device *parent_bdev, bool get_new)
{
	struct block_device_filter *table;
	struct block_device_filter *bdfilter;
	int idx;
	dev_t dev = parent_bdev->bd_dev;

	mutex_lock(&filter_lock);

	bdfilter = get_existing_filter(parent_bdev->bd_dev);
	if (bdfilter || !get_new)
		goto done;
	if (!stcvsm_bdget(dev)) {
		bdfilter = ERR_PTR(-ENOENT);
		goto fail1;
	}
	table = stcvsm_filter_table();
	for (idx = 0; idx < STCVSM_MAX_FILTERS; ++idx) {
		bdfilter = &table[idx];
		spin_lock(bdf_lock(bdfilter));
		if (atomic_read(&bdfilter->bdf_refs)) {
			spin_unlock(bdf_lock(bdfilter));
		} else {
			atomic_set(&bdfilter->bdf_refs, 1);
			atomic_set(&bdfilter->writes, 0);
			bdfilter->bdf_bdev = parent_bdev;
			bdfilter->bdf_disk = parent_bdev->bd_disk;
			spin_unlock(bdf_lock(bdfilter));

			stcvsm_info(LOG_FILTER, "(%d,%d) new %d\n", MINOR(dev), MINOR(dev), idx);
			mutex_unlock(&filter_lock);

			stcvsm_init_proc_filter(bdfilter);
			return bdfilter;
		}
	}

	bdfilter = ERR_PTR(-EINVAL);
	stcvsm_bdput(parent_bdev);
fail1:
	stcvsm_info(LOG_FILTER, "(%d,%d) error %ld\n", MAJOR(dev), MINOR(dev), PTR_ERR(bdfilter));
done:
	mutex_unlock(&filter_lock);
	return bdfilter;
}

void put_filter(struct block_device_filter *bdfilter)
{
	int ref = stcvsm_atomic_dec_and_lock(&bdfilter->bdf_refs, bdf_lock(bdfilter));

	if (ref) {
		struct block_device *parent_bdev = bdfilter->bdf_bdev;
		dev_t dev = bdfilter->bdf_bdev->bd_dev;

		/* We should never have num_snapshots or num_vdiffs if our refcount is zero */
		UASSERT(bdfilter->num_snapshots == 0, "Number of snapshots is not zero");
		UASSERT(bdfilter->num_vdiffs == 0, "Number of vdiffs is not zero");

		deactivate_filter(bdfilter);
		bdfilter->bdf_bdev = NULL;
		bdfilter->bdf_disk = NULL;
		spin_unlock(bdf_lock(bdfilter));

		stcvsm_remove_proc_filter(dev);
		stcvsm_bdput(parent_bdev);
	}
}

struct block_device_filter *get_filter(struct block_device *bdev, bool get_new)
{
	struct block_device_filter *retval;
	struct block_device *parent_bdev = get_parent_bdev(bdev);

	if (IS_ERR(parent_bdev)) {
		int err = PTR_ERR(parent_bdev);

		stcvsm_err(LOG_FILTER, "Failed to find parent block device (%d)\n", err);
		return ERR_PTR(err);
	}
	retval = get_parent_filter(parent_bdev, get_new);
	stcvsm_bdput(parent_bdev);
	return retval;
}
