/*
 * This source code (including its associated software) is owned by
 * StorageCraft Technology Corporation and is protected by United States and
 * international intellectual property law, including copyright laws, patent
 * laws, and treaty provisions. Use of the source code without modification in
 * connection with StorageCraft products is permitted subject to the End-User
 * License Agreement for StorageCraft Products ("EULA"). A copy of the
 * StorageCraft EULA is available here:
 * https://www.storagecraft.com/legal/license-agreement. Any other use,
 * reproduction, or disclosure without the written permission of StorageCraft
 * Technology Corporation is prohibited.
 *
 * © 2014-2019 StorageCraft Technology Corporation. All rights reserved.
 */

#include "debug.h"
#include "export.h"
#include "filter.h"
#include "import.h"
#include "vdiff.h"

#include "stcsnap_p.h"
#include "stcsnapdef_p.h"

static int copy_vdiff_stats(struct exported_state *state, struct block_device_filter **filters)
{
	int i;
	int err = 0;
	VSNAP_IOCTL_INFO info;
	mm_segment_t old_fs = get_fs();

	for (i = 0; i < state->num_vdiffs; i++) {
		dev_t dev = MKDEV(state->filters[i].major, state->filters[i].minor);
		struct vdiff_context *vdiff = get_filter_vdiff(filters[i], dev);

		info.major = state->filters[i].major;
		info.minor = state->filters[i].minor;
		set_fs(KERNEL_DS);
		err = state->stcvsm_ioctl_vdiff(dev, IOCTL_VSNAP_VDIFF_STATS_2, &info);
		set_fs(old_fs);
		if (err)
			break;

		if (info.Info.TotalSectors != vdiff->vd_sectors) {
			stcvsm_err(LOG_VDIFF, "Failed to migrate VDiff of (%d,%d) due to mismatched size\n",
					state->filters[i].major, state->filters[i].minor);
			return -EIO;
		}

		rean_vdiff(vdiff, info.Info.VDiffGuid, info.Info.VDiffFileGuid, info.Info.VDiffResetCount);
	}

	return err;
}

static void copy_vdiff_ranges(struct exported_state *state, struct block_device_filter **filters)
{
	int i;
	int err;
	mm_segment_t old_fs = get_fs();

	for (i = 0; i < state->num_vdiffs; i++) {
		dev_t dev = MKDEV(state->filters[i].major, state->filters[i].minor);
		struct vdiff_context *vdiff = get_filter_vdiff(filters[i], dev);
		VSNAP_VDIFF_GET_RANGE_OUTPUT_BUFFER buf;

		buf.major = state->filters[i].major;
		buf.minor = state->filters[i].minor;
		buf.start = 0;
		buf.NumberOfSectors = 0;

		while (1) {
			set_fs(KERNEL_DS);
			err = state->stcvsm_ioctl_vdiff(dev, IOCTL_VSNAP_VDIFF_GET_RANGE, &buf);
			set_fs(old_fs);

			if (err) {
				stcvsm_err(LOG_VDIFF, "Failed to migrate VDiff state; forcing differential on (%d,%d)\n",
						state->filters[i].major, state->filters[i].minor);
				vdiff->VDiffResetCount = 0;
				break;
			}
			if (buf.NumberOfSectors == 0)
				break;
			if (buf.AreDiffed)
				set_vdiff_range(vdiff, buf.start, buf.NumberOfSectors);
			buf.start += buf.NumberOfSectors;
		}
	}
}

int __init stcvsm_import_vdiff(void *(*stcvsm_export_vdiff)(unsigned int cmd))
{
	int i;
	int err;
	struct exported_state *state;
	struct block_device_filter **filters;

	/* start the transaction */
	state = stcvsm_export_vdiff(0);
	if (IS_ERR(state)) {
		err = PTR_ERR(state);
		stcvsm_err(LOG_VDIFF, "stcvsm_export_vdiff failed: %d\n", err);
		goto done;
	}
	if (state->reserved1 != 0 || state->reserved2 != 0) {
		err = -EINVAL;
		stcvsm_err(LOG_VDIFF, "incompatible flags: 0x%08x 0x%08x\n", state->reserved1,
				state->reserved2);
		goto cancel_trans;
	}
	filters = kcalloc(state->num_vdiffs, sizeof(struct block_device_filter *), GFP_KERNEL);
	if (filters == NULL) {
		err = -ENOMEM;
		goto cancel_trans;
	}

	/* initialize filters */
	for (i = 0; i < state->num_vdiffs; i++) {
		struct block_device *bdev;
		struct block_device_filter *bdfilter;
		struct vdiff_context *vdiff;
		dev_t dev = MKDEV(state->filters[i].major, state->filters[i].minor);

		bdev = stcvsm_open_blkdev(dev, FMODE_READ, NULL);
		if (IS_ERR(bdev)) {
			err = PTR_ERR(bdev);
			stcvsm_err(LOG_VDIFF, "stcvsm_open_blkdev failed: %d\n", err);
			goto stop_filters;
		}

		vdiff = alloc_vdiff(bdev);
		if (IS_ERR(vdiff)) {
			stcvsm_close_blkdev(bdev, FMODE_READ);
			err = PTR_ERR(vdiff);
			stcvsm_err(LOG_VDIFF, "alloc_vdiff failed: %d\n", err);
			goto stop_filters;
		}

		bdfilter = get_filter(bdev, true);
		stcvsm_close_blkdev(bdev, FMODE_READ);
		if (IS_ERR(bdfilter)) {
			free_vdiff(vdiff);
			err = PTR_ERR(bdfilter);
			stcvsm_err(LOG_VDIFF, "get_filter failed: %d\n", err);
			goto stop_filters;
		}
		err = attach_filter_vdiff(bdfilter, vdiff);
		put_filter(bdfilter);
		if (err) {
			free_vdiff(vdiff);
			stcvsm_err(LOG_VDIFF, "attach_filter_vdiff failed: %d\n", err);
			goto stop_filters;
		}
		filters[i] = bdfilter;
	}
	err = copy_vdiff_stats(state, filters);
	if (err) {
		stcvsm_err(LOG_VDIFF, "copy_vdiff_stats failed: %d\n", err);
		goto stop_filters;
	}

	/* divert IO flow; must commit the transaction past this point */
	for (i = 0; state->filters[i].major; i++) {
		/* activate filter, but the true orig_reqfn was exported */
		activate_filter(filters[i]);
		filters[i]->bdf_orig_reqfn = state->filters[i].orig_reqfn;
	}
	copy_vdiff_ranges(state, filters);

	/* commit; old driver can no longer restore reqfn */
	stcvsm_export_vdiff(1);

	kfree(state);
	kfree(filters);
	return 0;

stop_filters:
	while (i > 0) {
		dev_t dev;
		struct vdiff_context *vdiff;

		i--;
		dev = MKDEV(state->filters[i].major, state->filters[i].minor);
		vdiff = get_filter_vdiff(filters[i], dev);
		shutdown_vdiff(vdiff);
	}
	kfree(state);
	kfree(filters);
cancel_trans:
	stcvsm_export_vdiff(2);
done:
	return err;
}
