/*
 * This source code (including its associated software) is owned by
 * StorageCraft Technology Corporation and is protected by United States and
 * international intellectual property law, including copyright laws, patent
 * laws, and treaty provisions. Use of the source code without modification in
 * connection with StorageCraft products is permitted subject to the End-User
 * License Agreement for StorageCraft Products ("EULA"). A copy of the
 * StorageCraft EULA is available here:
 * https://www.storagecraft.com/legal/license-agreement. Any other use,
 * reproduction, or disclosure without the written permission of StorageCraft
 * Technology Corporation is prohibited.
 *
 * © 2014-2019 StorageCraft Technology Corporation. All rights reserved.
 */

#include "stcvsm_os.h"
#include "look_aside_list.h"


/*
 * This lookaside module is provided as an interface for memory allocation.
 *
 * It is intended as an abstract interface for a possible a future implementation of slab or mempool
 * allocation in replacement of the current simple kalloc/kfree implementation of the code below.
 *
 * An earlier attempt at using slab allocation caused an OOPS in the kernelk while the
 * stcvsm-regression test ran. It seems there is a lazy resource re-claiming code path in removing
 * entries in the /proc/slabinfo name space.  That slabinfo name space has a requirement that there
 * be NO duplicate names in the name space.  stcvsm-regression loads, unloads and reloads the module
 * hundreds of time during it course of operation and it appears a duplication name is added to the
 * name space while a prior instance name still persists in the symbol table.
 *
 * In additon,  stcvsm_linux module has a mechanism to be replaced with a newer version without
 * losing state.  It is possible, due to the lazy resource freeing, that using the slab allocation
 * could also cause an OOPS by module re-loading. Be aware of this for future consideration.
 */

void lookaside_initialize(struct lookaside_list *lookaside,
				void *allocate, void *free,
				u32 flags,
				size_t size,
				u32 tag,
				u16 depth)
{
	/* stop unreferenced variable warning */
	(void) allocate;
	(void) free;
	(void) flags;
	(void) tag;
	(void) depth;
	lookaside->pool = NULL;
	lookaside->name = NULL;
	lookaside->cache = NULL;
	lookaside->size = size;
}

void *lookaside_allocate(struct lookaside_list *lookaside)
{
	return kmalloc(lookaside->size, GFP_KERNEL);
}

void lookaside_free(struct lookaside_list *lookaside, void *entry)
{
	(void) lookaside;  /* stop unreferenced variable warning */
	kfree(entry);
}

void lookaside_delete(struct lookaside_list *lookaside)
{
	(void) lookaside;  /* stop unreferenced variable warning */
}
