/*
 * This source code (including its associated software) is owned by
 * StorageCraft Technology Corporation and is protected by United States and
 * international intellectual property law, including copyright laws, patent
 * laws, and treaty provisions. Use of the source code without modification in
 * connection with StorageCraft products is permitted subject to the End-User
 * License Agreement for StorageCraft Products ("EULA"). A copy of the
 * StorageCraft EULA is available here:
 * https://www.storagecraft.com/legal/license-agreement. Any other use,
 * reproduction, or disclosure without the written permission of StorageCraft
 * Technology Corporation is prohibited.
 *
 * © 2014-2019 StorageCraft Technology Corporation. All rights reserved.
 */

#include "debug.h"
#include "portability.h"

struct mutex stcvsm_freeze_lock;

/* NOTE:
 * bdget adds a reference count to bdev->bd_inode->i_count
 * bdput decrements i_count
 * blkdev_get does not increment i_count
 * blkdev_put decrements i_count
 *
 * blkdev_get adds a reference count bd_openers
 * blkdev_put decrements bd_openers
 *
 * while bd_openers is != 0, a reference count on the module that owns the block device is held,
 * preventing it from being unloadable.
 *
 * If i_count reaches 0, the inode is destroyed (along with the bdev), which will cause segfaults
 * in other programs if you decrement i_count more than you increment it.
 *
 * If bd_openers reaches 0, the reference count on the module is set to 0. If you do not take as
 * many bd_openers away as you take, you can cause other modules to be unloadable for no good
 * reason.
 *
 * Every bdget(stcvsm_bdget) requires a bdput(stcvsm_bdput) unless blkdev_get/put are called
 * (successfully or not).  Every successful blkdev_get(stcvsm_open_bdev) requires a
 * blkdev_put(stcvsm_close_blkdev).
 *
 * Or as <akpm@osdl.org> summarized:
 *   * bdget(): takes a number, returns a reference to block_device.
 *   * blkdev_get(): takes a reference to block_device and opens it.  If open
 *     fails - drops the reference to block_device passed to it.
 *   * blkdev_put(): takes a reference to block_device and closes it.  The
 *     reference is dropped.
 *   * bdput(): drops a reference to block_device.
 *   Note that behaviour of blkdev_get() and blkdev_put() is such that it makes
 *   for minimal cleanup code.
 */
struct block_device *stcvsm_bdget(dev_t dev)
{
	struct block_device *bdev;

	if (ERROR_INJECT(ERRINJ_BDGET))
		return NULL;

	bdev = bdget(dev);
	if (!bdev) {
		stcvsm_warning(LOG_MODULE, "Failed (%d,%d)\n", MAJOR(dev), MINOR(dev));
	} else {
		stcvsm_info(LOG_MODULE, "(%d,%d)(%d,%d)\n", MAJOR(dev), MINOR(dev),
				atomic_read(&bdev->bd_inode->i_count), bdev->bd_openers);
	}
	return bdev;
}

void stcvsm_bdput(struct block_device *bdev)
{
	if (bdev->bd_inode->i_state & I_CLEAR)
		return;
	stcvsm_info(LOG_MODULE, "(%d,%d)(%d,%d)\n", MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev),
		atomic_read(&bdev->bd_inode->i_count), bdev->bd_openers);
	bdput(bdev);
}

struct block_device *stcvsm_open_blkdev(dev_t dev, fmode_t mode, void *holder)
{
	int err;
	struct block_device *bdev;

	bdev = stcvsm_bdget(dev);
	if (bdev) {
		err = stcvsm_open_bdev(bdev, mode, holder);
		if (err == 0) {
			stcvsm_info(LOG_MODULE, "(%d,%d)(%d,%d)\n", MAJOR(dev), MINOR(dev),
					atomic_read(&bdev->bd_inode->i_count), bdev->bd_openers);
		} else {
			bdev = NULL;
		}
	} else {
		err = -ENOENT;
	}
	if (!bdev) {
		stcvsm_warning(LOG_MODULE, "Failed (%d,%d)\n", MAJOR(dev), MINOR(dev));
		return ERR_PTR(err);
	}
	return bdev;
}

int stcvsm_open_bdev(struct block_device *bdev, fmode_t mode, void *holder)
{
	int err;
	dev_t dev = bdev->bd_dev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 38)
	err = blkdev_get(bdev, mode, holder);
#else
	err = blkdev_get(bdev, mode & ~FMODE_EXCL);
	if (err == 0 && (mode & FMODE_EXCL) && holder) {
		if (bd_claim(bdev, holder) != 0) {
			err = -EBUSY;
			blkdev_put(bdev, mode);
		}
	}
#endif
	if (err < 0) {
		stcvsm_warning(LOG_MODULE, "Failed (%d,%d): %d\n", MAJOR(dev), MINOR(dev), -err);
		return err;
	}
	stcvsm_info(LOG_MODULE, "(%d,%d)(%d,%d)\n", MAJOR(dev), MINOR(dev),
		atomic_read(&bdev->bd_inode->i_count), bdev->bd_openers);
	return 0;
}

void stcvsm_close_blkdev(struct block_device *bdev, fmode_t mode)
{
	stcvsm_info(LOG_MODULE, "(%d,%d)(%d,%d)\n", MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev),
		atomic_read(&bdev->bd_inode->i_count), bdev->bd_openers);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 38)
	if (mode & FMODE_EXCL) {
		bd_release(bdev);
		mode &= ~FMODE_EXCL;
	}
#endif
	blkdev_put(bdev, mode);
}

int stcvsm_freeze_blkdev(struct block_device *bdev, struct super_block **sb)
{
	int err = 0;

	mutex_lock(&stcvsm_freeze_lock);
	stcvsm_info(LOG_MODULE, "Freezing (%d,%d)\n", MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));

	if (ERROR_INJECT(ERRINJ_BLKDEV_FREEZE))
		*sb = ERR_PTR(-EBUSY);
	else
		*sb = freeze_bdev(bdev);
	if (!*sb || IS_ERR(*sb)) {
		if (!*sb) {
			/* Even if there was no superblock to freeze, freeze_bdev incremented the
			 * bdev->bd_fsfreeze_count.  Not clear if this is a bug or by design.
			 * Thaw (which does error check for NULL sb) to fix the refcount.
			 *
			 * Note that this has a dangerous race window due to the kernel bug:  If
			 * another thread calls freeze_bdev before I thaw it here, then it sees
			 * bd_fsfreeze_count > 0, and assumes the superblock exists, and crashes.
			 */
			thaw_bdev(bdev, *sb);
		} else {
			err = PTR_ERR(*sb);
			stcvsm_warning(LOG_MODULE, "Failed (%d,%d): %d\n",
					MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev), -err);
			*sb = NULL;
		}
	} else {
		stcvsm_info(LOG_MODULE, "Frozen (%d,%d)\n",
				MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));
	}
	mutex_unlock(&stcvsm_freeze_lock);
	return err;
}

int stcvsm_thaw_blkdev(struct block_device *bdev, struct super_block *sb)
{
	/* As long as we always call stcvsm_freeze_blkdev, a null super_block should have already
	 * had thaw called.  Calling it a second time would mean that bd_fsfreeze_count > 0, and
	 * assumes the super_block exists.  See stcvsm_freeze_blkdev
	 */
	if (sb) {
		int err;

		stcvsm_info(LOG_MODULE, "Thawing (%d,%d)\n",
				MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));
		err = thaw_bdev(bdev, sb);
		if (err < 0) {
			stcvsm_warning(LOG_MODULE, "Failed (%d,%d): %d\n",
					MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev), -err);
			return err;
		}
		stcvsm_info(LOG_MODULE, "Thawed (%d,%d)\n",
				MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));
	} else {
		stcvsm_notice(LOG_MODULE, "Called with NULL super_block. (%d,%d)\n",
				MAJOR(bdev->bd_dev), MINOR(bdev->bd_dev));
	}
	return 0;
}
