/*
 * This source code (including its associated software) is owned by
 * StorageCraft Technology Corporation and is protected by United States and
 * international intellectual property law, including copyright laws, patent
 * laws, and treaty provisions. Use of the source code without modification in
 * connection with StorageCraft products is permitted subject to the End-User
 * License Agreement for StorageCraft Products ("EULA"). A copy of the
 * StorageCraft EULA is available here:
 * https://www.storagecraft.com/legal/license-agreement. Any other use,
 * reproduction, or disclosure without the written permission of StorageCraft
 * Technology Corporation is prohibited.
 *
 * © 2014-2019 StorageCraft Technology Corporation. All rights reserved.
 */

#ifndef STCVSM_PORTABILITY_H
#define STCVSM_PORTABILITY_H

#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/slab.h>
#include <linux/version.h>
#ifdef HAVE_BUFFER_HEAD_H
#include <linux/buffer_head.h>
#endif
#ifdef HAVE_BLK_TYPES_H
#include <linux/blk_types.h>
#endif

extern struct mutex stcvsm_freeze_lock;

int stcvsm_freeze_blkdev(struct block_device *bdev, struct super_block **sb);
int stcvsm_thaw_blkdev(struct block_device *bdev, struct super_block *sb);
struct block_device *stcvsm_bdget(dev_t dev);
struct block_device *stcvsm_open_blkdev(dev_t dev, fmode_t mode, void *holder);
int stcvsm_open_bdev(struct block_device *bdev, fmode_t mode, void *holder);
void stcvsm_bdput(struct block_device *bdev);
void stcvsm_close_blkdev(struct block_device *bdev, fmode_t mode);

#ifndef HAVE_PART_NR_SECTS_READ
#if BITS_PER_LONG==32
#error Need an atomic part_nr_sects_read()
#else
#define part_nr_sects_read(part) (part)->nr_sects
#endif
#endif

#ifdef HAVE_VOID_MAKE_REQUEST
#define MAKE_REQUEST_RETTYPE void
#undef MAKE_REQUEST_RETVAL
#elif defined HAVE_INT_MAKE_REQUEST
#define MAKE_REQUEST_RETTYPE int
#define MAKE_REQUEST_RETVAL 0
#else
#define MAKE_REQUEST_RETTYPE blk_qc_t
#define MAKE_REQUEST_RETVAL BLK_QC_T_NONE
#endif

/* k(un)map_atomic on 2.6.x required an explicit index; 3.x does automagic push/pop; 3.4 deprecated idx.
 * Be careful these calls could never be nested on old kernels to avoid stomping the same index.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 0)
#define stc_kmap_atomic(p) kmap_atomic(p, KM_USER0)
#define stc_kunmap_atomic(p) kunmap_atomic(p, KM_USER0)
#else
#define stc_kmap_atomic(p) kmap_atomic(p)
#define stc_kunmap_atomic(p) kunmap_atomic(p)
#endif

/* v3.2.51 a8203725 kmalloc_array introduced to check for integer overflow */
#ifndef HAVE_KMALLOC_ARRAY
static inline void *kmalloc_array(size_t n, size_t size, gfp_t flags)
{
	return kmalloc(n * size, flags);
}
#endif

static inline void stc_bio_set_op(struct bio *bio, int op)
{
	/* v4.8 1eff9d32 renamed bi_rw -> bi_opf */
#ifdef HAVE_BIO_BI_OPF
	bio->bi_opf |= op;
#else
	bio->bi_rw |= op;
#endif
}

static inline void stc_submit_bio(struct bio *bio)
{
	/* v4.8 4e49ea4a removed rw argument from submit_bio; expect bi_rw to be set instead */
#ifdef HAVE_SUBMIT_BIO_RW
	submit_bio(bio->bi_rw, bio);
#else
	submit_bio(bio);
#endif
}

static inline void stc_bio_endio(struct bio *bio, int error)
{
	/* v4.13 4e4cbee9 converted bi_error to blk_status_t bi_status */
	/* v4.8  a8203725 added a bi_error field */
#ifdef HAVE_BLK_STATUS_T
	// errno_to_blk_status will try to map a (negative) value from errno to blk_status_t
	switch (error) {
		case 0:
			bio->bi_status = BLK_STS_OK;
			break;
		case -EOPNOTSUPP:
			bio->bi_status = BLK_STS_NOTSUPP;
			break;
		default:
			bio->bi_status = BLK_STS_IOERR;

	}
	bio_endio(bio);
#elif defined HAVE_BIO_BI_ERROR
	bio->bi_error = error;
	bio_endio(bio);
#else
	bio_endio(bio, error);
#endif
}

#ifdef HAVE_BLK_STATUS_T
#define BIO_END_IO_PROTO(fn, b)    void fn(struct bio *b)
#define BIO_ERROR(bio)             (bio)->bi_status
#elif defined HAVE_BIO_BI_ERROR
#define BIO_END_IO_PROTO(fn, b)    void fn(struct bio *b)
#define BIO_ERROR(bio)             (bio)->bi_error
#else
#define BIO_END_IO_PROTO(fn, b)    void fn(struct bio *b, int bi_error)
#define BIO_ERROR(bio)             bi_error
#endif

#ifdef HAVE_IMMUTABLE_BIOVEC
typedef struct bvec_iter bio_iter_t;
typedef struct bio_vec bio_iter_bvec_t;
#define bio_sector(bio) ((bio)->bi_iter.bi_sector)
#define stc_kmap_atomic_bvec(bv) (stc_kmap_atomic((bv).bv_page) + (bv).bv_offset)
#else
typedef int bio_iter_t;
typedef struct bio_vec *bio_iter_bvec_t;
#define bio_sector(bio) ((bio)->bi_sector)
#define bio_iter_len(bio, iter) ((bio)->bi_io_vec[(iter)].bv_len)
#define stc_kmap_atomic_bvec(bv) (stc_kmap_atomic((bv)->bv_page) + (bv)->bv_offset)
#endif

/* BIOs used to have BIO_* flags and requests had REQ_* flags.  One could trigger another, but
 * they were not the same (BIO_FLUSH != REQ_FLUSH) which was very confusing.  Currently BIO_*
 * are removed and REQ_* is used everywhere.
 *
 * 2.6.32 RHEL	bio->bi_rw & BIO_FLUSH seems correct
 *		bio->bi_rw & BIO_RW_BARRIER exists but is deprecated
 *		bio->bi_rw & REQ_FLUSH exists in nvwe but perhaps wrong
 * <=2.6.35	bio->bi_rw & BIO_RW_BARRIER
 * 2.6.36	bio->bi_rw & REQ_HARDBARRIER
 * >=2.6.37	bio->bi_rw & REQ_FLUSH
 * v4.8 28a8f0d3 rename REQ_FLUSH to REQ_PREFLUSH
 */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 35)
#if LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 32) && defined(BIO_FLUSH)
#define STC_REQ_PREFLUSH BIO_FLUSH
#else
#define STC_REQ_PREFLUSH BIO_RW_BARRIER
#endif
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 36)
#define STC_REQ_PREFLUSH REQ_HARDBARRIER
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 8, 0)
#define STC_REQ_PREFLUSH REQ_FLUSH
#else
#define STC_REQ_PREFLUSH REQ_PREFLUSH
#endif

static inline int stc_bio_is_barrier(struct bio *bio)
{
#ifdef HAVE_BIO_BI_OPF
	return bio->bi_opf & STC_REQ_PREFLUSH;
#else
	return bio->bi_rw & STC_REQ_PREFLUSH;
#endif
}

static inline void stc_bio_assign_disk(struct bio *bio, struct block_device *bdev)
{
	if (!bio || !bdev)
		return;
#ifdef HAVE_BIO_GENDISK
	bio->bi_disk = bdev->bd_disk;
	bio->bi_partno = bdev->bd_partno;
#else
	bio->bi_bdev = bdev;
#endif
}

/* bio sets showed up around kernel version 3, and bio_clone was then aliased
 * to bio_clone_bioset using fs_bio_set. Around 4.15, bio_clone dissapeared,
 * but bio_clone_bioset stayed.
 * We get fs_bio_set from <linux/fs.h>
 */
static inline struct bio *stc_bio_clone(struct bio *bio, gfp_t gfp_mask) {
#ifdef HAVE_BIO_CLONE_BIOSET
	return bio_clone_bioset(bio, gfp_mask, fs_bio_set);
#else
	return bio_clone(bio, gfp_mask);
#endif
}

/* ext2 internals went private in 3.4.0.  Not yet used so just disable if missing. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 0)
#define WITH_EXT2BITMAP
#else
#undef WITH_EXT2BITMAP
#endif

/* Internal implementation details of proc_fs went private in 3.10.0.  Some vendors backport. */
#ifndef HAVE_PDE_DATA
#define PDE_DATA(inode) (PDE(inode)->data)
#endif

#endif
