/*
 * This source code (including its associated software) is owned by
 * StorageCraft Technology Corporation and is protected by United States and
 * international intellectual property law, including copyright laws, patent
 * laws, and treaty provisions. Use of the source code without modification in
 * connection with StorageCraft products is permitted subject to the End-User
 * License Agreement for StorageCraft Products ("EULA"). A copy of the
 * StorageCraft EULA is available here:
 * https://www.storagecraft.com/legal/license-agreement. Any other use,
 * reproduction, or disclosure without the written permission of StorageCraft
 * Technology Corporation is prohibited.
 *
 * © 2014-2019 StorageCraft Technology Corporation. All rights reserved.
 */

#include <linux/kthread.h>
#include <linux/slab.h>
#include <stdarg.h>

#include "debug.h"
#include "workqueue.h"

void wq_init(struct wq *wq, const char *name, ...)
{
	va_list args;

	spin_lock_init(&wq->lock);
	INIT_LIST_HEAD(&wq->work);
	init_waitqueue_head(&wq->more_work);
	wq->thread = NULL;
	va_start(args, name);
	wq->name = kvasprintf(GFP_KERNEL, name, args);
	va_end(args);
}

void wq_deinit(struct wq *wq)
{
	kfree(wq->name);
}

static int worker_thread(void *_wq)
{
	struct wq *wq = (struct wq *)_wq;
	DEFINE_WAIT(wait);

	while (1) {
		prepare_to_wait(&wq->more_work, &wait, TASK_INTERRUPTIBLE);
		if (!kthread_should_stop() && list_empty(&wq->work))
			schedule();
		finish_wait(&wq->more_work, &wait);

		spin_lock_irq(&wq->lock);
		while (!list_empty(&wq->work)) {
			struct wq_work *work = list_entry(wq->work.next, struct wq_work, list);

			list_del_init(wq->work.next);
			spin_unlock_irq(&wq->lock);

			work->fn(work);

			spin_lock_irq(&wq->lock);
		}
		spin_unlock_irq(&wq->lock);

		if (kthread_should_stop())
			break;
		if (!wq->thread) {
			if (wq->cleanup_fn)
				wq->cleanup_fn(wq->cleanup_data);
			break;
		}
	}
	return 0;
}

int wq_start(struct wq *wq)
{
	struct task_struct *p;

	if (ERROR_INJECT(ERRINJ_ALLOC))
		p = ERR_PTR(-ENOMEM);
	else
		p = kthread_create(worker_thread, wq, "%s", wq->name);
	if (IS_ERR(p))
		return PTR_ERR(p);
	wq->thread = p;
	wake_up_process(p);
	return 0;
}

int wq_stop(struct wq *wq, void (*cb)(void *data), void *data)
{
	int r = 0;

	if (wq->thread) {
		if (wq->thread == current) {
			wq->cleanup_fn = cb;
			wq->cleanup_data = data;
			wq->thread = NULL;
			r = -EAGAIN;
		} else {
			kthread_stop(wq->thread);  /* wakes too; no need to wake_up() myself */
			wq->thread = NULL;
			if (cb)
				cb(data);
		}
	}
	return r;
}

void wq_insert_work(struct wq *wq, struct wq_work *work)
{
	unsigned long lflags;

	INIT_LIST_HEAD(&work->list);
	spin_lock_irqsave(&wq->lock, lflags);
	list_add_tail(&work->list, &wq->work);
	wake_up(&wq->more_work);
	spin_unlock_irqrestore(&wq->lock, lflags);
}
